"""
    Test for routes functions : ensure responses are correct with mocked call to API
"""

from .resources import NemoResource
from .test_controller import NemoTestControllers
from flask_nemo import Nemo
from flask_nemo.plugins.default import Breadcrumb
from flask import Markup, Flask
from lxml import etree
from mock import Mock, patch, call
import MyCapytain
from jinja2.exceptions import TemplateNotFound


class NemoTestRoutes(NemoResource):
    """ Test Suite for Nemo
    """
    def test_route_index(self):
        """ Check that index return the template
        """
        self.assertEqual(self.nemo.r_index(), {"template": "main::index.html"})

    def test_route_collection(self):
        """ Test return values of route collection (list of textgroups
        """

        with patch('requests.get', return_value=self.getCapabilities) as patched:
            view = self.nemo.r_collection("latinLit")
            self.assertEqual(view["template"], "main::textgroups.html")
            self.assertEqual(len(view["textgroups"]), 3)
            self.assertIn("urn:cts:latinLit:phi1294", [str(textgroup.urn) for textgroup in view["textgroups"]])
            self.assertIsInstance(view["textgroups"][0], MyCapytain.resources.inventory.TextGroup)

    def test_route_texts(self):
        """ Test return values of route texts (list of texts for a textgroup
        """

        with patch('requests.get', return_value=self.getCapabilities) as patched:
            view = self.nemo.r_texts("latinLit", "phi1294")
            self.assertEqual(view["template"], "main::texts.html")
            self.assertEqual(len(view["texts"]), 2)
            self.assertEqual(
                sorted([str(view["texts"][0].urn), str(view["texts"][1].urn)]),
                sorted(["urn:cts:latinLit:phi1294.phi002.perseus-lat2", "urn:cts:latinLit:phi1294.phi002.perseus-eng2"])
            )
            self.assertIsInstance(view["texts"][0], MyCapytain.resources.inventory.Text)

    def test_route_version_chunker_replacement(self):
        """ Try to get valid reffs
        """

        urn = "urn:cts:latinLit:phi1294.phi002.perseus-lat2"
        def chunker(text, level):
            self.assertIsInstance(text, MyCapytain.resources.inventory.Text)
            self.assertEqual(str(text.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
            return True

        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc",
            chunker={"default": chunker}
        )

        with patch('requests.get', return_value=self.getValidReff) as patched:
            view = nemo.r_version("latinLit", "phi1294", "phi002", "perseus-lat2")
            self.assertIsInstance(view["version"], MyCapytain.resources.inventory.Text)
            patched.assert_called_once_with(
                NemoTestControllers.endpoint,
                params={
                    "inv": "annotsrc",
                    "request": "GetCapabilities"
                }
            )
            self.assertEqual(view["reffs"], True)

    def test_route_version_default_chunker(self):
        """ Try to get valid reffs
        """
        urn = "urn:cts:latinLit:phi1294.phi002.perseus-lat2"

        with patch('requests.get', return_value=self.getValidReff) as patched:
            view = self.nemo.r_version("latinLit", "phi1294", "phi002", "perseus-lat2")
            self.assertIsInstance(view["version"], MyCapytain.resources.inventory.Text)
            self.assertEqual(view["reffs"][0], ("1.pr.1", "1.pr.1"))

    def test_route_text_without_transform(self):
        """ Try to get valid reffs
        """
        urn = "urn:cts:latinLit:phi1294.phi002.perseus-lat2"

        with patch('requests.get', return_value=self.getValidReff) as patched:
            view = self.nemo.r_version("latinLit", "phi1294", "phi002", "perseus-lat2")
            self.assertIsInstance(view["version"], MyCapytain.resources.inventory.Text)
            self.assertEqual(view["reffs"][0], ("1.pr.1", "1.pr.1"))

    def test_route_passage_without_xslt(self):
        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc"
        )
        with patch('requests.get', return_value=self.getPassage_Route) as patched:
            view = self.nemo.r_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr.1")
            self.assertEqual(view["template"], "main::text.html")
            self.assertIsInstance(view["version"], MyCapytain.resources.inventory.Text)
            self.assertEqual(str(view["version"].urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
            self.assertEqual(view["prev"], "1.1.1")
            self.assertEqual(view["next"], "1.1.3")
            self.assertIsInstance(view["text_passage"], Markup)

            # Reparsing xml
            xml = etree.fromstring(str(view["text_passage"]))
            self.assertEqual(
                len(xml.xpath("//tei:body", namespaces={"tei":"http://www.tei-c.org/ns/1.0"})),
                1
            )
            self.assertEqual(
                len(xml.xpath("//tei:l", namespaces={"tei":"http://www.tei-c.org/ns/1.0"})),
                6
            )

    def test_route_passage_with_transform(self):
        """ Try with a non xslt just to be sure
        """
        urn = "urn:cts:latinLit:phi1294.phi002.perseus-lat2"
        def transformer(version, text):
            self.assertEqual(str(version.urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
            self.assertIsInstance(text, etree._Element)
            return "<a>Hello</a>"
        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc",
            transform={"default": transformer}
        )
        with patch('requests.get', return_value=self.getPassage_Route) as patched:
            view = nemo.r_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr.1")
            self.assertEqual(view["text_passage"], Markup("<a>Hello</a>"))

    def test_route_passage_with_xslt(self):
        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc",
            transform={"default": NemoTestControllers.body_xsl}
        )
        with patch('requests.get', return_value=self.getPassage_Route) as patched:
            view = nemo.r_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr.1")
            self.assertEqual(view["template"], "main::text.html")
            self.assertIsInstance(view["version"], MyCapytain.resources.inventory.Text)
            self.assertEqual(str(view["version"].urn), "urn:cts:latinLit:phi1294.phi002.perseus-lat2")
            self.assertEqual(view["prev"], "1.1.1")
            self.assertEqual(view["next"], "1.1.3")
            self.assertIsInstance(view["text_passage"], Markup)

            # Reparsing xml
            xml = etree.fromstring(str(view["text_passage"]))
            self.assertEqual(
                len(xml.xpath("//tei:notbody", namespaces={"tei":"http://www.tei-c.org/ns/1.0"})),
                1
            )

    def test_route_passage_with_urn_xslt(self):
        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc",
            transform={"urn:cts:latinLit:phi1294.phi002.perseus-lat2": NemoTestControllers.body_xsl}
        )
        with patch('requests.get', return_value=self.getPassage_Route) as patched:
            view = nemo.r_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr.1")
            # Reparsing xml
            xml = etree.fromstring(str(view["text_passage"]))
            self.assertEqual(
                len(xml.xpath("//tei:notbody", namespaces={"tei": "http://www.tei-c.org/ns/1.0"})),
                1
            )

    def test_route_passage_without_urn_xslt(self):
        nemo = Nemo(
            api_url=NemoTestControllers.endpoint,
            inventory="annotsrc",
            transform={"urn:cts:latinLit:phi1294.phi002.perseus-lat3": NemoTestControllers.body_xsl}
        )
        with patch('requests.get', return_value=self.getPassage_Route) as patched:
            view = nemo.r_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr.1")
            # Reparsing xml
            xml = etree.fromstring(str(view["text_passage"]))
            self.assertEqual(
                len(xml.xpath("//tei:body", namespaces={"tei": "http://www.tei-c.org/ns/1.0"})),
                1
            )

    def test_route_assets_404(self):
        with patch('flask_nemo.abort') as abort:
            self.nemo.r_assets("js", "wrong-js.js")
            abort.assert_called_once_with(404)

    def test_route_assets_all(self):
        nemo = Nemo(
            statics=["testing_data/getcapabilities.xml"],
            js=["testing_data/getcapabilities.xml"],
            css=["testing_data/style.css"]
        )
        nemo.blueprint = Mock()
        nemo.register_assets()
        with patch("flask_nemo.send_from_directory") as patched:
            nemo.r_assets("js", "getcapabilities.xml")
            nemo.r_assets("static", "getcapabilities.xml")
            nemo.r_assets("css", "style.css")
            patched.assert_has_calls([
                call(directory="testing_data", filename="getcapabilities.xml"),
                call(directory="testing_data", filename="getcapabilities.xml"),
                call(directory="testing_data", filename="style.css")
            ])

    def test_view_maker(self):
        """ View maker should take care of returning a lambda using the function self.route and the function
        identified by the parameter name
        """
        nemo = Nemo()
        nemo.route = Mock()

        view = nemo.view_maker("r_collection")
        view(collection="latinLit")
        nemo.route.assert_called_with(nemo.r_collection, collection="latinLit")

    def test_render_normal(self):
        """ Render adds informations, including url parameters in a url dict
        """
        with patch("requests.get", return_value=self.getCapabilities):
            with patch("flask_nemo.render_template") as patched:
                self.nemo.render("main::index.html", test="123", value="value", url={})
                patched.assert_called_once_with(
                    "main::index.html",
                    collections={'latinLit', 'greekLit'},
                    test="123",
                    value="value",
                    template="main::index.html",
                    lang="eng",
                    assets=self.nemo.assets,
                    url={},
                    breadcrumbs=[]
                )

    def test_render_textgroups(self):
        """ Render adds informations, including url parameters in a url dict
        """
        with patch("requests.get", return_value=self.getCapabilities):
            with patch("flask_nemo.render_template") as patched:
                data = self.nemo.render(
                    "index.html",
                    test="123",
                    value="value",
                    url={
                        "collection": "latinLit",
                        "textgroup": "phi1294"
                    })
                patched.assert_called_once_with(
                    "index.html",
                    collections={'latinLit', 'greekLit'},
                    test="123",
                    template="index.html",
                    value="value",
                    lang="eng",
                    assets=self.nemo.assets,
                    url={
                        "collection": "latinLit",
                        "textgroup": "phi1294",

                    },
                    breadcrumbs=[
                      {'link': '.r_collection', 'title': 'latinLit', 'args': {'collection': 'latinLit'}},
                      {'link': None, 'title': 'Martial', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit'}}
                    ],
                    textgroups=self.nemo.get_textgroups("latinLit")
                )

    def test_render_text(self):
        """ Render adds informations, including url parameters in a url dict
        """
        with patch("requests.get", return_value=self.getCapabilities):
            with patch("flask_nemo.render_template") as patched:
                self.nemo.render(
                    "index.html",
                    test="123",
                    value="value",
                    url={
                        "collection": "latinLit",
                        "textgroup": "phi1294",
                        "work": "phi002",
                        "text": "perseus-lat2"
                    })
                patched.assert_called_once_with(
                    "index.html",
                    template="index.html",
                    collections={'latinLit', 'greekLit'},
                    test="123",
                    value="value",
                    lang="eng",
                    assets=self.nemo.assets,
                    url={
                        "collection": "latinLit",
                        "textgroup": "phi1294",
                        "work": "phi002",
                        "text": "perseus-lat2"
                    },
                    breadcrumbs=[
                      {'link': '.r_collection', 'title': 'latinLit', 'args': {'collection': 'latinLit'}},
                      {'link': None, 'title': 'Martial', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit'}}
                    ],
                    textgroups=self.nemo.get_textgroups("latinLit"),
                    texts=self.nemo.get_texts("latinLit", "phi1294")
                )

    def test_route(self):
        """ nemo.route should apply fn and the args given
        """
        self.nemo.render = Mock()
        with patch("requests.get", return_value=self.getCapabilities):
            route = self.nemo.route(self.nemo.r_collection, collection="latinLit")
            self.nemo.render.assert_called_once_with(
                template="main::textgroups.html",
                textgroups=self.nemo.get_textgroups("latinLit"),
                url={"collection": "latinLit"}
            )

    def test_register_route(self):
        app = Flask(__name__)
        nemo = Nemo(app=app, base_url="/perseus")
        nemo.register()
        self.assertIn("flask_nemo", app.blueprints)

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/perseus/read/<collection>/<textgroup>/<work>/<version>/<passage_identifier>", [rule[0] for rule in rules])
        self.assertIn("flask_nemo.r_passage", [rule[1] for rule in rules])

        app = Flask(__name__)
        nemo = Nemo("nemo", app=app)
        nemo.register()
        self.assertIn("nemo", app.blueprints)

        rules = [(rule.rule, rule.endpoint) for rule in app.url_map.iter_rules()]
        self.assertIn("/nemo/read/<collection>/<textgroup>/<work>/<version>/<passage_identifier>", [rule[0] for rule in rules])
        self.assertIn("nemo.r_passage", [rule[1] for rule in rules])

        nemo = Nemo()
        self.assertEqual(nemo.register(), None)

    def test_additional_template(self):
        # Line 568-575
        app = Flask(__name__)
        nemo = Nemo(app=app, templates={"addendum": "tests/test_data/plugin_templates_main/main"})
        blueprint = nemo.create_blueprint()

        html, path, function = blueprint.jinja_loader.get_source("", "addendum::container.html")
        self.assertIn("I am A CONTAINER ! Isn't it sweet !", html)

        with self.assertRaises(TemplateNotFound):
            html, path, function = blueprint.jinja_loader.get_source("", "addendum::unknown.html")



# -*- coding: utf-8 -*-
"""
    Capitains Nemo
    ====

    Extensions for Flask to propose a Nemo extensions
"""


import os.path as op
import jinja2
from flask import render_template, Blueprint, abort, Markup, send_from_directory, Flask
import MyCapytain.retrievers.cts5
from MyCapytain.retrievers.proto import CTS as CtsProtoRetriever
import MyCapytain.resources.texts.tei
import MyCapytain.resources.texts.api
import MyCapytain.resources.inventory
from MyCapytain.common.reference import URN
from lxml import etree
from copy import deepcopy as copy
from pkg_resources import resource_filename
from collections import Callable, OrderedDict
import flask_nemo._data
import flask_nemo.filters
from flask_nemo.chunker import default_chunker as __default_chunker__, level_grouper as __level_grouper__
from flask_nemo.plugins.default import Breadcrumb
from flask_nemo.common import resource_qualifier, ASSETS_STRUCTURE


class Nemo(object):
    """ Nemo is an extension for Flask python micro-framework which provides
    a User Interface to your app for dealing with CTS API.

    :param app: Flask application
    :type app: Flask
    :param api_url: URL of the API Endpoint
    :type api_url: str
    :param retriever: CTS Retriever (Will be defaulted to api_url using cts5 retriever if necessary)
    :type retriever: MyCapytain.retrievers.proto.CTS
    :param base_url: Base URL to use when registering the endpoint (It cannot be "/" only !)
    :type base_url: str
    :param cache: SQLITE cache file name
    :type base_url: str
    :param expire: Time before expiration of cache, default 3600
    :type expire: int
    :param plugins: List of plugins to connect to the Nemo instance
    :type plugins: list(flask_nemo.plugin.PluginPrototype)
    :param template_folder: Folder in which the full set of main namespace templates can be found
    :type template_folder: str
    :param static_folder: Folder in which statics file can be found
    :type static_folder: str
    :param static_url_path: Base url to use for assets
    :type static_url_path: str
    :param urls: Function and routes to register (See Nemo.ROUTES)
    :type urls: [(str, str, [str])]
    :param inventory: Default inventory to use
    :type inventory: str
    :param transform: Dictionary of XSL filepath or transform function where default key is the default applied function
    :type transform: bool|dict
    :param urntransform: Dictionary of urn transform functions where default key is the default applied function
    :type urntransform: bool|dict
    :param chunker: Dictionary of function to group responses of GetValidReff
    :type chunker: {str: function(str, function(int))}
    :param prevnext: Dictionary of function to execute GetPrevNext
    :type prevnext: {str: function(str, function())}
    :param css: Path to additional stylesheets to load
    :type css: [str]
    :param js: Path to additional javascripts to load
    :type js: [str]
    :param templates: Register or override templates (Dictionary of namespace / directory containing template)
    :type templates: {str: str}
    :param statics: Path to additional statics such as picture to load
    :type statics: [str]
    :param prevent_plugin_clearing_assets: Prevent plugins to clear the static folder route
    :type prevent_plugin_clearing_assets: bool
    :param original_breadcrumb: Use the default Breadcrumb plugin packaged with Nemo (Default: True)
    :type original_breadcrumb: bool

    :ivar assets: Dictionary of assets loaded individually
    :ivar plugins: List of loaded plugins

    .. warning:: Until a C libxslt error is fixed ( https://bugzilla.gnome.org/show_bug.cgi?id=620102 ), it is not possible to use strip spaces in the xslt given to this application. See :ref:`lxml.strip-spaces`
    """

    ROUTES = [
        ("/", "r_index", ["GET"]),
        ("/read/<collection>", "r_collection", ["GET"]),
        ("/read/<collection>/<textgroup>", "r_texts", ["GET"]),
        ("/read/<collection>/<textgroup>/<work>/<version>", "r_version", ["GET"]),
        ("/read/<collection>/<textgroup>/<work>/<version>/<passage_identifier>", "r_passage", ["GET"])
    ]
    FILTERS = [
        "f_active_link",
        "f_collection_i18n",
        "f_formatting_passage_reference",
        "f_i18n_iso",
        "f_group_texts",
        "f_order_text_edition_translation",
        "f_hierarchical_passages",
        "f_is_str",
        "f_i18n_citation_type",
        "f_order_author",
        "f_annotation_filter"
    ]

    """ Assets dictionary model
    """
    ASSETS = copy(ASSETS_STRUCTURE)
    default_chunker = __default_chunker__

    def __init__(self, name=None, app=None, api_url="/", retriever=None, base_url="/nemo", cache=None, expire=3600,
                 plugins=None,
                 template_folder=None, static_folder=None, static_url_path=None,
                 urls=None, inventory=None, transform=None, urntransform=None, chunker=None, prevnext=None,
                 css=None, js=None, templates=None, statics=None,
                 prevent_plugin_clearing_assets=False,
                 original_breadcrumb=True):

        self.name = __name__
        if name:
            self.name = name
        self.prefix = base_url
        self.api_url = api_url

        if isinstance(retriever, CtsProtoRetriever):
            self.retriever = retriever
        else:
            self.retriever = MyCapytain.retrievers.cts5.CTS(self.api_url)

        if app is not None:
            self.app = app
        else:
            self.app = None

        self.api_inventory = inventory
        if self.api_inventory:
            self.retriever.inventory = self.api_inventory

        self.cache = None
        self.prevent_plugin_clearing_assets = prevent_plugin_clearing_assets

        if template_folder:
            self.template_folder = template_folder
        else:
            self.template_folder = resource_filename("flask_nemo", "data/templates")

        if static_folder:
            self.static_folder = static_folder
        else:
            self.static_folder = resource_filename("flask_nemo", "data/static")

        if static_url_path:
            self.static_url_path = static_url_path
        else:
            self.static_url_path = "/assets/nemo"
        self.blueprint = None

        if urls:
            self._urls = urls
        else:
            self._urls = copy(type(self).ROUTES)

        # Adding instance information
        self._urls = [tuple(list(url) + [None]) for url in self._urls]

        self._filters = copy(Nemo.FILTERS)
        self._filters = [tuple([filt] + [None]) for filt in self._filters]

        # Reusing self._inventory across requests
        self._inventory = None
        self.__transform = {
            "default": None
        }

        self.__urntransform = {
            "default": None
        }

        if isinstance(transform, dict):
            self.__transform.update(transform)

        if isinstance(urntransform, dict):
            self.__urntransform.update(urntransform)

        self.chunker = dict()
        self.chunker["default"] = type(self).default_chunker
        if isinstance(chunker, dict):
            self.chunker.update(chunker)

        self.prevnext = dict()
        self.prevnext["default"] = type(self).default_prevnext
        if isinstance(prevnext, dict):
            self.prevnext.update(prevnext)

        # Setting up assets
        self.__assets__ = copy(type(self).ASSETS)
        if css and isinstance(css, list):
            for css_s in css:
                filename, directory = resource_qualifier(css_s)
                self.__assets__["css"][filename] = directory
        if js and isinstance(js, list):
            for javascript in js:
                filename, directory = resource_qualifier(javascript)
                self.__assets__["js"][filename] = directory
        if statics and isinstance(statics, list):
            for static in statics:
                directory, filename = op.split(static)
                self.__assets__["static"][filename] = directory

        self.__plugins_render_views__ = []
        self.__plugins__ = OrderedDict()
        if original_breadcrumb:
            self.__plugins__["nemo.breadcrumb"] = Breadcrumb(name="breadcrumb")
        if isinstance(plugins, list):
            for plugin in plugins:
                self.__plugins__[plugin.name] = plugin

        self.__templates_namespaces__ = [
            ("main", self.template_folder)
        ]
        self.__instance_templates__ = []

        if isinstance(templates, dict):
            self.__instance_templates__.extend(
                [(namespace, folder) for namespace, folder in templates.items()]
            )
        self.__template_loader__ = dict()

        if app:
            self.init_app(self.app)

    @property
    def plugins(self):
        return self.__plugins__

    @property
    def assets(self):
        return self.__assets__

    def init_app(self, app=None):
        """ Initiate the application

        :param app: Flask application on which to add the extension
        :type app: flask.Flask
        """
        # Legacy code
        if "CTS_API_URL" in app.config:
            self.api_url = app.config['CTS_API_URL']
        if "CTS_API_INVENTORY" in app.config:
            self.api_inventory = app.config['CTS_API_INVENTORY']
        if app:
            self.app = app

        self.register()

    def transform(self, work, xml):
        """ Transform input according to potentiallyregistered XSLT

        .. note:: Due to XSLT not being able to be used twice, we rexsltise the xml at every call of xslt
        .. warning:: Until a C libxslt error is fixed ( https://bugzilla.gnome.org/show_bug.cgi?id=620102 ), it is not possible to use strip tags in the xslt given to this application

        :param work: Work object containing metadata about the xml
        :type work: MyCapytains.resources.inventory.Text
        :param xml: XML to transform
        :type xml: etree._Element
        :return: String representation of transformed resource
        :rtype: str
        """
        # We check first that we don't have
        if str(work.urn) in self.__transform:
            func = self.__transform[str(work.urn)]
        else:
            func = self.__transform["default"]

        # If we have a string, it means we get a XSL filepath
        if isinstance(func, str):
            with open(func) as f:
                xslt = etree.XSLT(etree.parse(f))
            return etree.tostring(xslt(xml), encoding=str, method="html", xml_declaration=None, pretty_print=False, with_tail=True, standalone=None)

        # If we have a function, it means we return the result of the function
        elif isinstance(func, Callable):
            return func(work, xml)
        # If we have None, it meants we just give back the xml
        elif func is None:
            return etree.tostring(xml, encoding=str)

    def transform_urn(self, urn):
        """ Transform urn according to configurable function

        :param urn: URN to transform
        :type urn: URN
        :return: the URN (transformed or not)
        :rtype: URN
        """
        # We check first that we don't have an override function
        # N.B. overrides will be on the text level, not the passage
        if urn.upTo(URN.NO_PASSAGE) in self.__urntransform:
            func = self.__urntransform[urn.upTo(URN.NO_PASSAGE)]
        else:
            func = self.__urntransform["default"]

        # If we have a function, it means we return the result of the function
        if isinstance(func, Callable):
            return func(urn)
        # If we have None, it means we just give back the urn as string
        return urn

    def get_inventory(self):
        """ Request the api endpoint to retrieve information about the inventory

        :return: The text inventory
        :rtype: MyCapytain.resources.inventory.TextInventory
        """
        if self._inventory:
            return self._inventory

        reply = self.retriever.getCapabilities(inventory=self.api_inventory)
        inventory = MyCapytain.resources.inventory.TextInventory(resource=reply)
        self._inventory = inventory
        return self._inventory

    def get_collections(self):
        """ Filter inventory and make a list of available collections

        :return: A set of CTS Namespaces
        :rtype: set(str)
        """
        inventory = self.get_inventory()
        urns = set(
            [inventory.textgroups[textgroup].urn.namespace for textgroup in inventory.textgroups]
        )
        return urns

    def get_textgroups(self, collection_urn=None):
        """ Retrieve textgroups

        :param collection_urn: Collection to use for filtering the textgroups
        :type collection_urn: str
        :return: List of textgroup filtered by collection
        :rtype: [MyCapytain.resources.inventory.Textgroup]
        """
        inventory = self.get_inventory()
        if collection_urn is not None:
            return Nemo.map_urns(inventory, collection_urn, 2, "textgroups")
        return list(inventory.textgroups.values())

    def get_works(self, collection_urn=None, textgroup_urn=None):
        """ Retrieve works

        :param collection_urn: Collection to use for filtering the textgroups
        :type collection_urn: str
        :param textgroup_urn: Textgroup to use for filtering the works
        :type textgroup_urn: str
        :return: List of work filtered by collection/Textgroup
        :rtype: [MyCapytain.resources.inventory.Work]
        """
        if collection_urn is not None and textgroup_urn is not None:
            textgroup = list(
                filter(lambda x: Nemo.filter_urn(x, 3, textgroup_urn), self.get_textgroups(collection_urn))
            )
            if len(textgroup) == 1:
                return textgroup[0].works.values()
            else:
                return []
        elif collection_urn is None and textgroup_urn is None:
            return [work for textgroup in self.get_inventory().textgroups.values() for work in textgroup.works.values()]
        else:
            raise ValueError("Get_Work takes either two None value or two set up value")

    def get_texts(self, collection_urn=None, textgroup_urn=None, work_urn=None):
        """ Retrieve texts

        :param collection_urn: Collection to use for filtering the textgroups
        :type collection_urn: str
        :param textgroup_urn: Textgroup to use for filtering the works
        :type textgroup_urn: str
        :param work_urn: Work to use for filtering the texts
        :type work_urn: str
        :return: List of texts filtered by parameters
        :rtype: [MyCapytain.resources.inventory.Text]
        """
        if collection_urn is not None and textgroup_urn is not None and work_urn is not None:
            work = list(
                filter(lambda x: Nemo.filter_urn(x, 4, work_urn), self.get_works(collection_urn, textgroup_urn))
            )
            if len(work) == 1:
                return work[0].texts.values()
            else:
                return []
        elif collection_urn is not None and textgroup_urn is not None and work_urn is None:
            return [
                text
                for work in self.get_works(collection_urn, textgroup_urn)
                for text in work.texts.values()
            ]
        elif collection_urn is None and textgroup_urn is None and work_urn is None:
            return [
                text
                for textgroup in self.get_inventory().textgroups.values()
                for work in textgroup.works.values()
                for text in work.texts.values()
            ]
        else:
            raise ValueError("Get_Work takes either two None value or two set up value")

    def get_text(self, collection_urn, textgroup_urn, work_urn, version_urn):
        """ Retrieve one version of a Text

        :param collection_urn: Collection to use for filtering the textgroups
        :type collection_urn: str
        :param textgroup_urn: Textgroup to use for filtering the works
        :type textgroup_urn: str
        :param work_urn: Work identifier to use for filtering texts
        :type work_urn: str
        :param version_urn: Version identifier
        :type version_urn: str
        :return: A Text represented by the various parameters
        :rtype: MyCapytain.resources.inventory.Text
        """
        work = list(
            filter(lambda x: Nemo.filter_urn(x, 4, work_urn), self.get_works(collection_urn, textgroup_urn))
        )
        if len(work) == 1:
            texts = work[0].texts.values()
        else:
            texts = []

        texts = [text for text in texts if text.urn.version == version_urn]
        if len(texts) == 1:
            return texts[0]
        abort(404)

    def get_reffs(self, collection, textgroup, work, version):
        """ Get the setup for valid reffs.

        Returns the inventory text object with its metadata and a callback function taking a level parameter
        and returning a list of strings.

        :param collection: Collection identifier
        :type collection: str
        :param textgroup: Textgroup identifier
        :type textgroup: str
        :param work: Work identifier
        :type work: str
        :param version: Version identifier
        :type version: str
        :return: Text with its metadata, callback function to retrieve validreffs
        :rtype: (MyCapytains.resources.texts.api.Text, lambda: [str])
        """
        text = self.get_text(collection, textgroup, work, version)
        reffs = MyCapytain.resources.texts.api.Text(
            "urn:cts:{0}:{1}.{2}.{3}".format(collection, textgroup, work, version),
            self.retriever,
            citation=text.citation
        )
        return text, lambda level: reffs.getValidReff(level=level)

    def get_passage(self, collection, textgroup, work, version, passage_identifier):
        """ Retrieve the passage identified by the parameters


        :param collection: Collection identifier
        :type collection: str
        :param textgroup: Textgroup identifier
        :type textgroup: str
        :param work: Work identifier
        :type work: str
        :param version: Version identifier
        :type version: str
        :param passage_identifier: Reference Identifier
        :type passage_identifier: str
        :return: A Passage object containing informations about the passage
        :rtype: MyCapytain.resources.texts.api.Passage
        """
        text = MyCapytain.resources.texts.api.Text(
            "urn:cts:{0}:{1}.{2}.{3}".format(collection, textgroup, work, version),
            self.retriever
        )
        passage = text.getPassage(passage_identifier)
        return passage

    def r_index(self):
        """ Homepage route function

        :return: Template to use for Home page
        :rtype: {str: str}
        """
        return {"template": "main::index.html"}

    def r_collection(self, collection):
        """ Collection content browsing route function

        :param collection: Collection identifier
        :type collection: str
        :return: Template and textgroups contained in given collections
        :rtype: {str: Any}
        """
        return {
            "template": "main::textgroups.html",
            "textgroups": self.get_textgroups(collection)
        }

    def r_texts(self, collection, textgroup):
        """ Textgroup content browsing route function

        :param collection: Collection identifier
        :type collection: str
        :param textgroup: Textgroup Identifier
        :type textgroup: str
        :return: Template and texts contained in given textgroup
        :rtype: {str: Any}
        """
        return {
            "template": "main::texts.html",
            "texts": self.get_texts(collection, textgroup)
        }

    def r_version(self, collection, textgroup, work, version):
        """ Text exemplar references browsing route function

        :param collection: Collection identifier
        :type collection: str
        :param textgroup: Textgroup Identifier
        :type textgroup: str
        :param work: Work identifier
        :type work: str
        :param version: Version identifier
        :type version: str
        :return: Template, version inventory object and references urn parts
        :rtype: {
            "template" : str,
            "version": MyCapytains.resources.inventory.Text,
            "reffs": [str]
            }
        """
        version, reffs = self.get_reffs(collection, textgroup, work, version)
        reffs = self.chunk(version, reffs)
        return {
            "template": "main::version.html",
            "version": version,
            "reffs": reffs
        }

    def r_passage(self, collection, textgroup, work, version, passage_identifier):
        """ Retrieve the text of the passage

        :param collection: Collection identifier
        :type collection: str
        :param textgroup: Textgroup Identifier
        :type textgroup: str
        :param work: Work identifier
        :type work: str
        :param version: Version identifier
        :type version: str
        :param passage_identifier: Reference identifier
        :type passage_identifier: str
        :return: Template, version inventory object and Markup object representing the text
        :rtype: {str: Any}
        """
        edition = self.get_text(collection, textgroup, work, version)
        text = self.get_passage(collection, textgroup, work, version, passage_identifier)

        passage = self.transform(edition, text.xml)
        prev, next = self.getprevnext(text, Nemo.prevnext_callback_generator(text))
        urn = self.transform_urn(text.urn)
        return {
            "template": "main::text.html",
            "version": edition,
            "text_passage": Markup(passage),
            "urn": urn,
            "prev": prev,
            "next": next
        }

    def r_assets(self, filetype, asset):
        """ Route for specific assets.

        :param asset: Filename of an asset
        :return: Response
        """
        if filetype in self.assets and asset in self.assets[filetype] and self.assets[filetype][asset]:
            return send_from_directory(
                directory=self.assets[filetype][asset],
                filename=asset
            )
        abort(404)

    def register_assets(self):
        """ Merge and register assets, both as routes and dictionary

        :return: None
        """
        self.blueprint.add_url_rule(
            # Register another path to ensure assets compatibility
            "{0}.secondary/<filetype>/<asset>".format(self.static_url_path),
            view_func=self.r_assets,
            endpoint="secondary_assets",
            methods=["GET"]
        )

    def create_blueprint(self):
        """ Create blueprint and register rules

        :return: Blueprint of the current nemo app
        :rtype: flask.Blueprint
        """
        self.register_plugins()

        self.blueprint = Blueprint(
            self.name,
            "nemo",
            url_prefix=self.prefix,
            template_folder=self.template_folder,
            static_folder=self.static_folder,
            static_url_path=self.static_url_path
        )

        for url, name, methods, instance in self._urls:
            self.blueprint.add_url_rule(
                url,
                view_func=self.view_maker(name, instance),
                endpoint=_plugin_endpoint_rename(name, instance),
                methods=methods
            )

        self.register_assets()
        self.register_filters()

        # We extend the loading list by the instance value
        self.__templates_namespaces__.extend(self.__instance_templates__)
        # We generate a template loader
        for namespace, directory in self.__templates_namespaces__[::-1]:
            if namespace not in self.__template_loader__:
                self.__template_loader__[namespace] = []
            self.__template_loader__[namespace].append(
                jinja2.FileSystemLoader(op.abspath(directory))
            )
        self.blueprint.jinja_loader = jinja2.PrefixLoader(
            {namespace: jinja2.ChoiceLoader(paths) for namespace, paths in self.__template_loader__.items()},
            "::"
        )

        return self.blueprint

    def view_maker(self, name, instance=None):
        """ Create a view

        :param name: Name of the route function to use for the view.
        :type name: str
        :return: Route function which makes use of Nemo context (such as menu informations)
        :rtype: function
        """
        if instance:
            return lambda **kwargs: self.route(getattr(instance, name), **kwargs)
        else:
            return lambda **kwargs: self.route(getattr(self, name), **kwargs)

    def render(self, template, **kwargs):
        """ Render a route template and adds information to this route.

        :param template: Template name.
        :type template: str
        :param kwargs: dictionary of named arguments used to be passed to the template
        :type kwargs: dict
        :return: Http Response with rendered template
        :rtype: flask.Response
        """

        kwargs["collections"] = self.get_collections()
        kwargs["lang"] = "eng"

        if Nemo.in_and_not_in("textgroup", "textgroups", kwargs):
            kwargs["textgroups"] = self.get_textgroups(kwargs["url"]["collection"])

            if Nemo.in_and_not_in("text", "texts", kwargs):
                kwargs["texts"] = self.get_texts(kwargs["url"]["collection"], kwargs["url"]["textgroup"])

        kwargs["assets"] = self.assets
        kwargs["template"] = template

        for plugin in self.__plugins_render_views__:
            kwargs.update(plugin.render(**kwargs))

        return render_template(kwargs["template"], **kwargs)

    def route(self, fn, **kwargs):
        """ Route helper : apply fn function but keep the calling object, *ie* kwargs, for other functions

        :param fn: Function to run the route with
        :type fn: function
        :param kwargs: Parsed url arguments
        :type kwargs: dict
        :return: HTTP Response with rendered template
        :rtype: flask.Response
        """
        new_kwargs = fn(**kwargs)

        # If there is no templates, we assume that the response is finalized :
        if not isinstance(new_kwargs, dict):
            return new_kwargs

        new_kwargs["url"] = kwargs
        return self.render(**new_kwargs)

    def register(self):
        """ Register the app using Blueprint

        :return: Nemo blueprint
        :rtype: flask.Blueprint
        """
        if self.app is not None:
            if not self.blueprint:
                self.blueprint = self.create_blueprint()
            self.app.register_blueprint(self.blueprint)
            return self.blueprint
        return None

    def register_filters(self):
        """ Register filters for Jinja to use

       .. note::  Extends the dictionary filters of jinja_env using self._filters list
        """
        for _filter, instance in self._filters:
            if not instance:
                self.app.jinja_env.filters[
                    _filter.replace("f_", "")
                ] = getattr(flask_nemo.filters, _filter)
            else:
                self.app.jinja_env.filters[
                    _filter.replace("f_", "")
                ] = getattr(instance, _filter.replace("_{}".format(instance.name), ""))

    def register_plugins(self):
        """ Register plugins in Nemo instance

        - Clear routes first if asked by one plugin
        - Clear assets if asked by one plugin and replace by the last plugin registered static_folder
        - Register each plugin
            - Append plugin routes to registered routes
            - Append plugin filters to registered filters
            - Append templates directory to given namespaces
            - Append assets (CSS, JS, statics) to given resources 
            - Append render view (if exists) to Nemo.render stack
        """
        if len([plugin for plugin in self.__plugins__.values() if plugin.clear_routes]) > 0:  # Clear current routes
            self._urls = list()

        clear_assets = [plugin for plugin in self.__plugins__.values() if plugin.clear_assets]
        if len(clear_assets) > 0 and not self.prevent_plugin_clearing_assets:  # Clear current Assets
            self.__assets__ = copy(type(self).ASSETS)
            static_path = [plugin.static_folder for plugin in clear_assets if plugin.static_folder]
            if len(static_path) > 0:
                self.static_folder = static_path[-1]

        for plugin in self.__plugins__.values():
            self._urls.extend([(url, function, methods, plugin) for url, function, methods in plugin.routes])
            self._filters.extend([(filt, plugin) for filt in plugin.filters])
            self.__templates_namespaces__.extend(
                [(namespace, directory) for namespace, directory in plugin.templates.items()]
            )
            for asset_type in self.__assets__:
                for key, value in plugin.assets[asset_type].items():
                    self.__assets__[asset_type][key] = value
            if plugin.augment:
                self.__plugins_render_views__.append(plugin)
            plugin.register_nemo(self)

    def chunk(self, text, reffs):
        """ Handle a list of references depending on the text identifier using the chunker dictionary.

        :param text: Text object from which comes the references
        :type text: MyCapytains.resources.texts.api.Text
        :param reffs: Callback function to retrieve a list of string with a level parameter
        :type reffs: callback(level)
        :return: Transformed list of references
        :rtype: [str]
        """
        if str(text.urn) in self.chunker:
            return self.chunker[str(text.urn)](text, reffs)
        return self.chunker["default"](text, reffs)

    def getprevnext(self, passage, callback):
        """ Retrieve previous and next passage using

        :param text: Text object from which comes the references
        :type text: MyCapytains.resources.texts.api.Passage
        :param reffs: Callback function to retrieve a tuple where first element is the previous passage, and second the next
        :type reffs: callback()
        :return: Reference of previous passage, reference of next passage
        :rtype: (str, str)
        """
        if str(passage.urn) in self.prevnext:
            return self.prevnext[str(passage.urn)](passage, callback)
        return self.prevnext["default"](passage, callback)

    @staticmethod
    def in_and_not_in(identifier, collection, kwargs):
        """ Check if an element identified by identifier is in kwargs but not the collection containing it

        :param identifier: URL Identifier of one kind of element (Textgroup, work, etc.)
        :type identifier: str
        :param collection: Resource identifier of one kind of element (Textgroup, work, etc.)
        :type collection: str
        :param kwargs: Arguments passed to a template
        :type kwargs: {str: Any}
        :return: Indicator of presence of required informations
        :rtype: bool
        """
        return identifier in kwargs["url"] and collection not in kwargs

    @staticmethod
    def default_prevnext(passage, callback):
        """ Default deliver of prevnext informations

        :param passage: Passage for which to get previous and following reference
        :type passage: MyCapytains.resources.texts.api.Passage
        :param callback: Function to retrieve those information
        :type callback: function

        :return: Tuple representing previous and following reference
        :rtype: (str, str)
        """
        previous, following = passage.prev, passage.next

        if previous is None and following is None:
            previous, following = callback()

        if previous is not None:
            previous = str(previous.reference)
        if following is not None:
            following = str(following.reference)
        return previous, following

    @staticmethod
    def prevnext_callback_generator(passage):
        """ Default callback generator to retrieve prev and next value of a passage

        :param passage: Passage for which to get previous and following reference
        :type passage: MyCapytains.resources.texts.api.Passage
        :return: Function to retrieve those information
        :rtype: function
        """
        def callback():
            return MyCapytain.resources.texts.api.Passage.prevnext(
                passage.parent.resource.getPrevNextUrn(urn=str(passage.urn))
            )
        return callback

    @staticmethod
    def map_urns(items, query, part_of_urn=1, attr="textgroups"):
        """ Small function to map urns to filter out a list of items or on a parent item

        :param items: Inventory object
        :type items: MyCapytains.resources.inventory.Resource
        :param query: Part of urn to check against
        :type query: str
        :param part_of_urn: Identifier of the part of the urn
        :type part_of_urn: int
        :return: Items corresponding to the object children filtered by the query
        :rtype: list(items.children)
        """

        if attr is not None:
            return [
                item
                for item in getattr(items, attr).values()
                if Nemo.filter_urn(item, part_of_urn, query)
            ]

    @staticmethod
    def filter_urn(item, part_of_urn, query):
        """ Small function to map urns to filter out a list of items or on a parent item

        :param item: Inventory object
        :param query: Part of urn to check against
        :type query: str
        :param part_of_urn: Identifier of the part of the urn
        :type part_of_urn: int
        :return: Items corresponding to the object children filtered by the query
        :rtype: list(items.children)
        """

        return str(item.urn.__getattribute__([
            "", "urn_namespace", "namespace", "textgroup", "work", "version", "reference"
        ][part_of_urn]).lower()) == query.lower().strip()


def _plugin_endpoint_rename(fn_name, instance):
    """ Rename endpoint function name to avoid conflict when namespacing is set to true

    :param fn_name: Name of the route function
    :param instance: Instance bound to the function
    :return: Name of the new namespaced function name
    """

    if instance and instance.namespaced:
        fn_name = "r_{0}_{1}".format(instance.name, fn_name[2:])
    return fn_name


def cmd():
    import argparse
    parser = argparse.ArgumentParser(description='Capitains Nemo CTS UI')
    parser.add_argument('endpoint', metavar='endpoint', type=str,
                       help='CTS API Endpoint')
    parser.add_argument('--port', type=int, default=8000,
                       help='Port to use for the HTTP Server')
    parser.add_argument('--host', type=str, default="127.0.0.1",
                       help='Host to use for the HTTP Server')
    parser.add_argument('--inventory', type=str, default=None,
                       help='Inventory to request from the endpoint')
    parser.add_argument('--css', type=str, default=None, nargs='*',
                       help='Full path to secondary css file')
    parser.add_argument('--groupby', type=int, default=25,
                       help='Number of passage to group in the deepest level of the hierarchy')
    parser.add_argument('--debug', action="store_true", default=False, help="Set-up the application for debugging")

    args = parser.parse_args()

    if args.endpoint:
        app = Flask(
            __name__
        )
        # We set up Nemo
        nemo = Nemo(
            app=app,
            name="nemo",
            base_url="",
            css=args.css,
            inventory=args.inventory,
            api_url=args.endpoint,
            chunker={"default": lambda x, y: __level_grouper__(x, y, groupby=args.groupby)}
        )

        # We run the app
        app.debug = args.debug
        app.run(port=args.port, host=args.host)

if __name__ == "__main__":
    cmd()



"""
    Test the Nautilus endpoint with the app.test_client
"""


from unittest import TestCase
from .resources import NautilusDummy
from flask_nemo import Nemo
from flask_nemo.chunker import level_grouper
from flask import Flask, jsonify


class NemoTestRoutes(TestCase):
    """ Test Suite for Nemo
    """
    def setUp(self):
        app = Flask("Nemo")
        app.debug = True
        nemo = Nemo(
            app=app,
            base_url="",
            retriever=NautilusDummy,
            chunker={"default": lambda x, y: level_grouper(x, y, groupby=30)},
            css=[
                "./tests/test_data/empty.css",
                "//foo.bar/test.css",
                "http://bar.foo/test.css",
                "https://super.secure/mypasswordin.css"
            ],
            js=[
                "./tests/test_data/empty.js",
                "//foo.bar/test.js",
                "http://bar.foo/test.js",
                "https://super.secure/mypasswordin.js"
            ],
        )

        self.client = app.test_client()

    def test_default_template_assets(self):
        """ Test that the index menu is correctly built """
        query_data = str(self.client.get("/").data)

        jss = [
            "//foo.bar/test.js",
            "http://bar.foo/test.js",
            "https://super.secure/mypasswordin.js"
        ]
        assert_length_js = len(jss)
        for js in jss:
            self.assertIn(
                '<script src="{0}"></script>'.format(js), query_data,
                "Templates should correctly link external js"
            )
            assert_length_js -= 1
        self.assertEqual(assert_length_js, 0, "All js file should have been checked")

        csss = [
            "//foo.bar/test.css",
            "http://bar.foo/test.css",
            "https://super.secure/mypasswordin.css"
        ]
        assert_length_js = len(csss)
        for css in csss:
            self.assertIn(
                '<link rel="stylesheet" href="{0}">'.format(css), query_data,
                "Templates should correctly link external css"
            )
            assert_length_js -= 1
        self.assertEqual(assert_length_js, 0, "All css files should have been checked")

        self.assertIn(
            '<link rel="stylesheet" href="/assets/nemo.secondary/css/empty.css">', query_data,
            "Secondary Internal CSS should be linked"
        )
        self.assertIn(
            '<script src="/assets/nemo.secondary/js/empty.js"></script>', query_data,
            "Secondary Internal JS should be retrieved"
        )

    def test_serving_secondary_assets(self):
        """ Ensure secondary assets are served
        """
        query_data = str(self.client.get("/assets/nemo.secondary/css/empty.css").data)
        self.assertIn(".empty {", query_data, "Local CSS File should be read")

        query_data = str(self.client.get("/assets/nemo.secondary/js/empty.js").data)
        self.assertIn("var empty = True;", query_data, "Local JS File should be read")

    def test_serving_primary_assets(self):
        """ Ensure primary assets are served
        """
        query_data = str(self.client.get("/assets/nemo/css/theme.min.css").data)
        self.assertIn("body, html", query_data, "Primary Assets should be served")

    def test_serving_overwritten_primary_assets(self):
        """ Ensure primary assets are served
        """
        app = Flask("Nemo")
        app.debug = True
        nemo = Nemo(
            app=app,
            base_url="",
            retriever=NautilusDummy,
            chunker={"default": lambda x, y: level_grouper(x, y, groupby=30)},
            static_folder="tests/test_data/assets"
        )

        client = app.test_client()
        query_data = str(client.get("/assets/nemo/fake.png").data)
        self.assertIn("fake", query_data, "Primary Assets should be served when overwritting")

        self.assertEqual(
            client.get("/assets/nemo/css/theme.min.css").status_code, 404,
            "Old Primary Assets should not be served when overwritting"
        )

    def test_index_menu(self):
        """ Test that the index menu is correctly built """
        query_data = str(self.client.get("/").data)
        self.assertIn(
            '<a class="" href="/read/farsiLit">farsiLit</a>', query_data,
            "App should have link to farsiLit through local repository-endpoint object"
        )
        self.assertIn(
            '<a class="" href="/read/latinLit">Latin</a>', query_data,
            "App should have link to latinLit through local repository-endpoint object"
        )

    def test_namespace_page(self):
        """ Test that the namespace page has correct informations : """
        query_data = str(self.client.get("/read/latinLit").data)
        self.assertIn(
            '<a class="" href="/read/farsiLit">farsiLit</a>', query_data,
            "App should have link to farsiLit through local repository-endpoint object"
        )
        self.assertIn(
            '<li><a class="" href="/read/latinLit/phi1294">Martial</a></li>', query_data,
            "App should have link to authors through local repository-endpoint object"
        )

    def test_author_page(self):
        """ Test that author page contains what is relevant : editions and translations """
        query_data = str(self.client.get("/read/latinLit/phi1294").data)
        self.assertIn(
            '<a class="" href="/read/farsiLit">farsiLit</a>', query_data,
            "App should have link to farsiLit through local repository-endpoint object"
        )
        self.assertIn(
            '<a href="/read/latinLit/phi1294/phi002/perseus-lat2">', query_data,
            "App should have link to the text object"
        )

    def test_text_page(self):
        """ Test that text page contains what is relevant : passages"""
        query_data = str(self.client.get("/read/latinLit/phi1294/phi002/perseus-lat2").data)
        self.assertIn(
            '<a class="" href="/read/farsiLit">farsiLit</a>', query_data,
            "App should have link to farsiLit through local repository-endpoint object"
        )
        self.assertIn(
            '<a href="/read/latinLit/phi1294/phi002/perseus-lat2/1.pr.1-1.pr.22">', query_data,
            "App should have link to farsiLit through local repository-endpoint object"
        )

    def test_passage_page(self):
        """ Test that passage page contains what is relevant : text and next passages"""
        query_data = str(self.client.get("/read/latinLit/phi1294/phi002/perseus-lat2/1.pr.1-1.pr.22").data)
        self.assertIn(
            'Marsus, sic Pedo, sic Gaetulicus, sic quicumque perlegi', query_data,
            "Text should be visible"
        )
        self.assertIn(
            'href="/read/latinLit/phi1294/phi002/perseus-lat2/1.1.1-1.3.8"', query_data,
            "App should have link to the next passage"
        )

    def test_json_route(self):
        """ Test that adding routes to Nemo instance (without plugin) which output json works
        """
        import json
        test_data = {"SomeDict": "IsGood"}

        class NemoJson(Nemo):
            ROUTES = Nemo.ROUTES + [("/getJson", "r_json", ["GET"])]

            def r_json(self):
                """ Route with no templates should return none as first value
                """
                return jsonify(test_data)

        app = Flask("Nemo")
        app.debug = True
        nemo = NemoJson(app=app, base_url="", retriever=NautilusDummy)
        client = app.test_client()
        query_data = json.loads(client.get("/getJson").data.decode('utf8'))

        self.assertEqual(
            query_data, test_data,
            "Original Dict and Decoded Output JSON should be equal"
        )

    def test_breadcrumb(self):
        """ Ensure breadcrumb is bydefault loaded
        """
        query_data = str(self.client.get("/read/latinLit/phi1294").data)
        self.assertIn(
            '<li><a href="/read/latinLit">latinLit</a></li>', query_data,
            "Breadcrumb should be visible"
        )
        self.assertIn(
            '<li class="active">Martial</li>', query_data,
            "Breadcrumb should be visible"
        )

    def test_no_default_breadcrumb(self):
        """ Ensure breadcrumb is bydefault loaded
        """

        app = Flask("Nemo")
        app.debug = True
        nemo = Nemo(app=app, base_url="", retriever=NautilusDummy, original_breadcrumb=False)
        client = app.test_client()
        query_data = str(client.get("/read/latinLit/phi1294").data)
        self.assertNotIn(
            '<ol class="breadcrumb">', query_data,
            "Breadcrumb should not be visible"
        )
        self.assertNotIn(
            '<li class="active">Martial</li>', query_data,
            "Breadcrumb should not be visible"
        )



"""

Test file for filters.

"""

from flask import Flask
from flask_nemo import Nemo
from flask_nemo.filters import f_active_link, f_collection_i18n, f_formatting_passage_reference, f_group_texts, \
    f_hierarchical_passages, f_i18n_citation_type, f_i18n_iso, f_is_str, f_order_author,\
    f_order_text_edition_translation, f_annotation_filter
import MyCapytain
from .resources import NemoResource


class TestFilters(NemoResource):

    def test_f_order_author(self):
        """ Test ordering authors
        """
        a = MyCapytain.resources.inventory.TextGroup()
        b = MyCapytain.resources.inventory.TextGroup()
        c = MyCapytain.resources.inventory.TextGroup()
        d = MyCapytain.resources.inventory.TextGroup()
        e = MyCapytain.resources.inventory.TextGroup()
        f = MyCapytain.resources.inventory.TextGroup(
            urn="urn:cts:latinLit:phi1294.phi002.perseus-lat2"
        )
        a.metadata["groupname"]["eng"] = "Martial"
        b.metadata["groupname"]["eng"] = "Virgil"
        c.metadata["groupname"]["eng"] = "Cicero"
        d.metadata["groupname"]["eng"] = "Augustine"
        e.metadata["groupname"]["eng"] = "Suetone"
        f.metadata["groupname"]["eng"] = None
        sort = f_order_author([a, b, c, d, e, f])
        self.assertEqual(
            sort,
            [d, c, a, e, b, f]
        )

    def test_f_active_link(self):
        """ Test checking if something is in the path
        """
        wrong = f_active_link("phi003", {"collection": "latinLit"})
        wrong2 = f_active_link("phi003", {"collection": "latinLit", "textgroup": "phi0003"})
        right = f_active_link("phi003", {"collection": "latinLit", "textgroup": "phi0003", "work": "phi003"})
        self.assertEqual(wrong, "")
        self.assertEqual(wrong2, "")
        self.assertEqual(right, "active")

    def test_f_collection_i18n(self):
        """ Test internationalization of collection identifier
        """
        self.assertEqual(f_collection_i18n("latinLit"), "Latin")
        self.assertEqual(f_collection_i18n("greekLit"), "Ancient Greek", "Should use known namespace information")
        self.assertEqual(f_collection_i18n("afrLit"), "Afrikaans", "Should match regular expression and use data")
        self.assertEqual(f_collection_i18n("aaaLit"), "aaaLit", "Should keep the string when unknown to parser")

    def test_f_formatting_passage_reference(self):
        """ Test split of passage range identifier
        """
        self.assertEqual(f_formatting_passage_reference("1.1-1.2"), "1.1")
        self.assertEqual(f_formatting_passage_reference("1.1"), "1.1")

    def test_f_i18n_iso(self):
        """ Test split of passage range identifier
        """
        self.assertEqual(f_i18n_iso("eng"), "English")
        self.assertEqual(f_i18n_iso("eng", "fre"), "anglais")
        self.assertEqual(f_i18n_iso("eng", "ger"), "English")

    def test_f_order_text_edition_translation(self):
        """ Check the reordering filter
        """
        Text = MyCapytain.resources.inventory.Text
        a, b, c, d = Text(subtype="Translation"), Text(subtype="Edition"), Text(subtype="Edition"), Text(subtype="Translation")
        self.assertEqual(f_order_text_edition_translation([a,b,c,d]), [b, c, a, d])

    def test_register_filter(self):
        app = Flask(__name__)
        self.nemo = Nemo(app=app)
        self.nemo.register_filters()
        self.assertEqual(self.nemo.app.jinja_env.filters["formatting_passage_reference"], f_formatting_passage_reference)
        self.assertEqual(self.nemo.app.jinja_env.filters["collection_i18n"], f_collection_i18n)
        self.assertEqual(self.nemo.app.jinja_env.filters["active_link"], f_active_link)

    def test_f_hierarchical_passages(self):
        """Test for normal, simple passage hierarchical conversion
        :return:
        """
        reffs = [("1.5.8", "Line 8"), ("1.5.9", "Line 9"), ("1.6.8", "Line 7"), ("2.5.8", "Line 12")]
        citation_line = MyCapytain.common.reference.Citation(name="line")
        citation_poem = MyCapytain.common.reference.Citation(name="poem", child=citation_line)
        citation_book = MyCapytain.common.reference.Citation(name="book", child=citation_poem)
        text = MyCapytain.resources.inventory.Text()
        text.citation = citation_book
        converted = f_hierarchical_passages(reffs, text)
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 8"], "1.5.8")
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 9"], "1.5.9")
        self.assertEqual(converted["%book|1%"]["%poem|6%"]["Line 7"], "1.6.8")
        self.assertEqual(converted["%book|2%"]["%poem|5%"]["Line 12"], "2.5.8")
        self.assertEqual(len(converted), 2)
        self.assertEqual(len(converted["%book|1%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|5%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|6%"]), 1)
        self.assertEqual(len(converted["%book|2%"]), 1)

    def test_f_hierarchical_range_passages(self):
        """Test for range passage hierarchical conversion
        :return:
        """
        reffs = [("1.5.8-1.5.9", "Line 8"), ("1.5.9-1.5.15", "Line 9"), ("1.6.8-2.9.16", "Line 7"), ("2.5.8-16.45.928", "Line 12")]
        citation_line = MyCapytain.common.reference.Citation(name="line")
        citation_poem = MyCapytain.common.reference.Citation(name="poem", child=citation_line)
        citation_book = MyCapytain.common.reference.Citation(name="book", child=citation_poem)
        text = MyCapytain.resources.inventory.Text()
        text.citation = citation_book
        converted = f_hierarchical_passages(reffs, text)
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 8"], "1.5.8-1.5.9")
        self.assertEqual(converted["%book|1%"]["%poem|5%"]["Line 9"], "1.5.9-1.5.15")
        self.assertEqual(converted["%book|1%"]["%poem|6%"]["Line 7"], "1.6.8-2.9.16")
        self.assertEqual(converted["%book|2%"]["%poem|5%"]["Line 12"], "2.5.8-16.45.928")
        self.assertEqual(len(converted), 2)
        self.assertEqual(len(converted["%book|1%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|5%"]), 2)
        self.assertEqual(len(converted["%book|1%"]["%poem|6%"]), 1)
        self.assertEqual(len(converted["%book|2%"]), 1)

    def test_f_is_str(self):
        """ Test string
        """
        self.assertEqual(f_is_str("h"), True)
        self.assertEqual(f_is_str([]), False)

    def test_f_i18n_citation_type(self):
        self.assertEqual(f_i18n_citation_type("%book|1%"), "Book 1")

    def test_annotation_filter(self):
        class Annotation(object):
            def __init__(self, type_uri):
                self.type_uri = type_uri

        a = Annotation("1")
        b = Annotation("2")
        c = Annotation("3")
        d = Annotation("2")

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "1", 1), a,
            "We should get only one of type 1"
        )

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "4", 1), None,
            "We should get nothing"
        )

        self.assertEqual(
            f_annotation_filter([a, b, c, d], "2", 2), d,
            "We should get the second item"
        )


from MyCapytain.resources.texts.api import Text
from MyCapytain.common.reference import URN
from flask_nemo.query.proto import QueryPrototype
from flask_nemo.query.annotation import AnnotationResource
from werkzeug.exceptions import NotFound


class SimpleQuery(QueryPrototype):
    """ Query Interface for hardcoded annotations.

    :param annotations: List of tuple of (CTS URN Targeted, URI of the Annotation, Type of the annotation) or/and AnnotationResources
    :type annotations: [(str, str, str) or AnnotationResource]
    :param resolver: Resolver
    :type resolver: Resolver

    This interface requires to be connected to Nemo upon instantiation to expand annotations :

    >>> nemo = Nemo("/", endpoint="http://cts.perseids.org")
    >>> query = SimpleQuery([...])
    >>> query.process(nemo)

    """

    def __init__(self, annotations, resolver=None):
        super(SimpleQuery, self).__init__(None)
        self.__annotations__ = []
        self.__nemo__ = None
        self.__resolver__ = resolver

        for resource in annotations:
            if isinstance(resource, tuple):
                target, body, type_uri = resource
                self.__annotations__.append(AnnotationResource(
                    body, target, type_uri, self.__resolver__
                ))
            else:
                self.__annotations__.append(resource)

    def process(self, nemo):
        """ Register nemo and parses annotations

        .. note:: Process parses the annotation and extends informations about the target URNs by retrieving resource in range

        :param nemo: Nemo
        """
        self.__nemo__ = nemo
        for annotation in self.__annotations__:
            text = self.__getText__(annotation.target.urn)
            if annotation.target.urn.reference.end \
                or len(annotation.target.urn.reference.list) < len(text.citation):
                annotation.target.expanded = frozenset(self.__getinnerreffs__(
                    text=text,
                    urn=annotation.target.urn
                ))
            else:
                annotation.target.expanded = frozenset([str(annotation.target.urn)])

    def __getText__(self, urn):
        """ Return a metadata text object

        :param urn: URN object of the passage
        :return: Text
        """
        return self.__nemo__.get_text(
            urn.namespace,
            urn.textgroup,
            urn.work,
            urn.version
        )

    @property
    def annotations(self):
        return self.__annotations__

    def getResource(self, sha):
        try:
            return [annotation for annotation in self.annotations if annotation.sha == sha][0]
        except IndexError:
            raise NotFound

    def getAnnotations(self,
            urns=None,
            wildcard=".", include=None, exclude=None,
            limit=None, start=1,
            expand=False,
            **kwargs
        ):
        annotations = []

        if not urns:
            return len(self.annotations), sorted(self.annotations, key=lambda x: x.uri)

        if not isinstance(urns, list):
            urns = [urns]

        for urn in urns:

            if not isinstance(urn, URN):
                _urn = URN(urn)
            else:
                _urn = urn

            if _urn.reference and _urn.reference.end:
                urns_in_range = self.__getinnerreffs__(
                    text=self.__getText__(_urn),
                    urn=_urn
                )
            else:
                urns_in_range = [str(urn)]
            urns_in_range = frozenset(urns_in_range)
            annotations.extend([
                annotation
                for annotation in self.annotations
                if str(_urn) == str(annotation.target.urn) or  # Exact Match
                bool(urns_in_range.intersection(annotation.target.expanded))  # Deeper than
            ])

        annotations = list(set(annotations))

        return len(annotations), sorted(annotations, key=lambda x: x.uri)

    def __getinnerreffs__(self, text, urn):
        """ Resolve the list of urns between in a range

        :param urn: Urn of the passage
        :type urn: URN
        :return:
        """
        text = Text(
            str(text.urn),
            self.__nemo__.retriever,
            citation=text.citation
        )
        return text.getValidReff(reference=urn.reference, level=len(text.citation))


from ..resources import NemoResource
from mock import patch
from flask_nemo.plugins.default import Breadcrumb


class TestBreadcrumb(NemoResource):
    """ Test Breadcrumb default Plugins
    """
    def test_make_passage_breadcrumb(self):
        """ passage breadcrumb should include all components up to passage and passage not linked
        """
        with patch("requests.get", return_value=self.getCapabilities):
            make_breadcrumbs = Breadcrumb().render
            bc = make_breadcrumbs(
                textgroups=self.nemo.get_textgroups(),
                version=self.nemo.get_text("latinLit", "phi1294", "phi002", "perseus-lat2"),
                lang="eng",
                url={
                    "collection": "latinLit",
                    "textgroup": "phi1294",
                    "work": "phi002",
                    "version": "perseus-lat2",
                    "passage_identifier": "1.1"
                    })["breadcrumbs"]
            self.assertEqual(bc,[
                {'link': '.r_collection', 'title': 'latinLit', 'args': {'collection': 'latinLit'}},
                {'link': '.r_texts', 'title': 'Martial', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit'}},
                {'link': '.r_version', 'title': 'Epigrammata Label', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit', 'work':'phi002','version':'perseus-lat2'}},
                {'link': None, 'title': '1.1', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit', 'work':'phi002','version':'perseus-lat2', 'passage_identifier':"1.1"}}
            ])

    def test_make_version_breadcrumb(self):
        """ version breadcrumb should include all components up to version and version not linked
        """
        with patch("requests.get", return_value=self.getCapabilities):
            make_breadcrumbs = Breadcrumb().render
            bc = make_breadcrumbs(
                textgroups=self.nemo.get_textgroups(),
                version = self.nemo.get_text("latinLit","phi1294","phi002","perseus-lat2"),
                lang="eng",
                url={
                    "collection": "latinLit",
                    "textgroup": "phi1294",
                    "work": "phi002",
                    "version": "perseus-lat2"
                    })["breadcrumbs"]
            self.assertEqual(bc,[
                {'link': '.r_collection', 'title': 'latinLit', 'args': {'collection': 'latinLit'}},
                {'link': '.r_texts', 'title': 'Martial', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit'}},
                {'link': None, 'title': 'Epigrammata Label', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit', 'work':'phi002','version':'perseus-lat2'}}
            ])

    def test_make_textgroup_breadcrumb(self):
        """ textgroup breadcrumb should include all components up to textgroup and textgroup not linked
        """
        with patch("requests.get", return_value=self.getCapabilities):
            make_breadcrumbs = Breadcrumb().render
            bc = make_breadcrumbs(
                textgroups=self.nemo.get_textgroups(),
                lang="eng",
                url={
                    "collection": "latinLit",
                    "textgroup": "phi1294",
                    })["breadcrumbs"]
            self.assertEqual(bc, [
                {'link': '.r_collection', 'title': 'latinLit', 'args': {'collection': 'latinLit'}},
                {'link': None, 'title': 'Martial', 'args': {'textgroup': 'phi1294', 'collection': 'latinLit'}}
            ])

    def test_make_collection_breadcrumb(self):
        """ collection breadcrumb should include only collection not linked
        """
        with patch("requests.get", return_value=self.getCapabilities):
            make_breadcrumbs = Breadcrumb().render
            bc = make_breadcrumbs(
                lang="eng",
                url={"collection": "latinLit"})["breadcrumbs"]
            self.assertEqual(bc, [
                {'link': None, 'title': 'latinLit', 'args': {'collection': 'latinLit'}}
            ])



import flask_nemo._data
from flask_nemo.common import regMatch, getFromDict
from collections import defaultdict, OrderedDict


def f_order_author(textgroups, lang="eng"):
    """ Order a list of textgroups

    :param textgroups: list of textgroups to be sorted
    :param lang: Language to display
    :return: Sorted list
    """
    __textgroups__ = {
        tg.metadata["groupname"][lang] or str(tg.urn): tg
        for tg in textgroups
    }

    return [
       __textgroups__[key]
       for key in sorted(list(__textgroups__.keys()))
   ]


def f_active_link(string, url):
    """ Check if current string is in the list of names

    :param string: String to check for in url
    :return: CSS class "active" if valid
    :rtype: str
    """
    if string in url.values():
        return "active"
    return ""


def f_collection_i18n(string, lang="eng"):
    """ Return a i18n human readable version of a CTS domain such as latinLit

    :param string: CTS Domain identifier
    :type string: str
    :return: Human i18n readable version of the CTS Domain
    :rtype: str
    """
    if string in flask_nemo._data.COLLECTIONS:
        return flask_nemo._data.COLLECTIONS[string]
    elif regMatch.match(string):
        lg = string[0:3]
        if lg in flask_nemo._data.ISOCODES and lang in flask_nemo._data.ISOCODES[lg]:
            return flask_nemo._data.ISOCODES[lg][lang]
    return string


def f_formatting_passage_reference(string):
    """ Get the first part only of a two parts reference

    :param string: A urn reference part
    :type string: str
    :return: First part only of the two parts reference
    :rtype: str
    """
    return string.split("-")[0]


def f_i18n_iso(isocode, lang="eng"):
    """ Replace isocode by its language equivalent

    :param isocode: Three character long language code
    :param lang: Lang in which to return the language name
    :return: Full Text Language Name
    """
    if lang not in flask_nemo._data.AVAILABLE_TRANSLATIONS:
        lang = "eng"

    try:
        return flask_nemo._data.ISOCODES[isocode][lang]
    except KeyError:
        return "Unknown"


def f_group_texts(versions_list):
    """ Takes a list of versions and regroup them by work identifier

    :param versions_list: List of text versions
    :type versions_list: [Text]
    :return: List of texts grouped by work
    :rtype: [(Work, [Text])]
    """
    works = {}
    texts = defaultdict(list)
    for version in versions_list:
        if version.urn.work not in works:
            works[version.urn.work] = version.parents[0]
        texts[version.urn.work].append(version)
    return [
        (works[index], texts[index])
        for index in works
    ]


def f_order_text_edition_translation(versions_list):
    """ Takes a list of versions and put translations after editions

    :param versions_list: List of text versions
    :type versions_list: [Text]
    :return: List where first members will be editions
    :rtype: [Text]
    """
    translations = []
    editions = []
    for version in versions_list:
        if version.subtype == "Translation":
            translations.append(version)
        else:
            editions.append(version)
    return editions + translations


def f_hierarchical_passages(reffs, version):
    """ A function to construct a hierarchical dictionary representing the different citation layers of a text

    :param reffs: passage references with human-readable equivalent
    :type reffs: [(str, str)]
    :param version: text from which the reference comes
    :type version: MyCapytain.resources.inventory.Text
    :return: nested dictionary representing where keys represent the names of the levels and the final values represent the passage reference
    :rtype: OrderedDict
    """
    d = OrderedDict()
    levels = [x for x in version.citation]
    for cit, name in reffs:
        ref = cit.split('-')[0]
        levs = ['%{}|{}%'.format(levels[i].name, v) for i, v in enumerate(ref.split('.'))]
        getFromDict(d, levs[:-1])[name] = cit
    return d


def f_is_str(value):
    """ Check if object is a string

    :param value: object to check against
    :return: Return if value is a string
    """
    return isinstance(value, str)


def f_i18n_citation_type(string, lang="eng"):
    """ Take a string of form %citation_type|passage% and format it for human

    :param string: String of formation %citation_type|passage%
    :param lang: Language to translate to
    :return: Human Readable string

    .. note :: To Do : Use i18n tools and provide real i18n
    """
    s = " ".join(string.strip("%").split("|"))
    return s.capitalize()


def f_annotation_filter(annotations, type_uri, number):
    """ Annotation filtering filter

    :param annotations: List of annotations
    :type annotations: [AnnotationResource]
    :param type_uri: URI Type on which to filter
    :type type_uri: str
    :param number: Number of the annotation to return
    :type number: int
    :return: Annotation(s) matching the request
    :rtype: [AnnotationResource] or AnnotationResource
    """
    filtered = [
        annotation
        for annotation in annotations
        if annotation.type_uri == type_uri
    ]
    number = min([len(filtered), number])
    if number == 0:
        return None
    else:
        return filtered[number-1]


# -*- coding: utf-8 -*-

from flask_nemo.plugin import PluginPrototype
from pkg_resources import resource_filename


class Breadcrumb(PluginPrototype):
    """
        The Breadcrumb plugin is enabled by default in Nemo.
        It can be overwritten or removed. It simply adds a breadcrumb

    """
    HAS_AUGMENT_RENDER = True
    TEMPLATES = {"main": resource_filename("flask_nemo", "data/templates_plugins/breadcrumb")}

    def render(self, **kwargs):
        """ Make breadcrumbs for a route

        :param kwargs: dictionary of named arguments used to construct the view
        :type kwargs: dict
        :return: List of dict items the view can use to construct the link.
        :rtype: {str: list({ "link": str, "title", str, "args", dict})}
        """
        breadcrumbs = []
        # this is the list of items we want to accumulate in the breadcrumb trail.
        # item[0] is the key into the kwargs["url"] object and item[1] is the  name of the route
        # setting a route name to None means that it's needed to construct the route of the next item in the list
        # but shouldn't be included in the list itself (this is currently the case for work --
        # at some point we probably should include work in the navigation)
        crumbtypes = [
            ["collection", ".r_collection"],
            ["textgroup", ".r_texts"],
            ["work", None],
            ["version", ".r_version"],
            ["passage_identifier", ".r_passage"]
        ]
        for idx, crumb_type in enumerate(crumbtypes):
            if kwargs["url"] and crumb_type[0] in kwargs["url"]:
                crumb = {}
                # what we want to display as the crumb title depends upon what it is
                # in the future, having a common display_name property on the model would be helpful to avoid
                # this logic here
                if crumb_type[0] == "textgroup":
                    # get the groupname of the current textgroup
                    item = list(filter(lambda textgroup: textgroup.urn.textgroup == kwargs["url"]["textgroup"], kwargs["textgroups"]))
                    crumb["title"] = item[0].metadata["groupname"][kwargs["lang"]]
                elif crumb_type[0] == "version":
                    # get the label of the current version
                    crumb["title"] = kwargs["version"].metadata["label"][kwargs["lang"]]
                else:
                    # for everything else, just use the value as metadata isn't applicable
                    crumb["title"] = kwargs["url"][crumb_type[0]]
                # iterate through the crumb types and pull together the args that lead up to this type
                # so that we can reconstruct the route to just this part of the breadcrumb trail
                crumb_args = {}
                iteridx = idx

                while iteridx >= 0:
                    crumb_args[crumbtypes[iteridx][0]] = kwargs["url"][crumbtypes[iteridx][0]]
                    iteridx -= 1

                crumb["link"] = crumb_type[1]
                crumb["args"] = crumb_args
                # skip items in the trail that are only used to construct others
                if crumb_type[1]:
                    breadcrumbs.append(crumb)

        # don't link the last item in the trail
        if len(breadcrumbs) > 0:
            breadcrumbs[-1]["link"] = None

        return {"breadcrumbs": breadcrumbs}



import unittest
from flask.ext.nemo import Nemo
from flask import Markup, Flask
from capitains_nautilus.mycapytain import NautilusRetriever
import logging


def create_test_app(debug=False, config=None):
    app = Flask(__name__)
    app.debug = debug

    if config:
        app.config.update(**config)
    return app


class RequestPatch(object):
    """ Request patch object to deal with patching reply in flask.ext.nemo
    """
    def __init__(self, f):
        self.__text = f.read()

    @property
    def text(self):
        return self.__text


class RequestPatchChained(object):
    """ Request patch object to deal with patching reply in flask.ext.nemo
    """
    def __init__(self, requests):
        self.resource = [other.text for other in requests]

    @property
    def text(self):
        return self.resource.pop(0)


class NemoResource(unittest.TestCase):
    """ Test Suite for Nemo
    """
    endpoint = "http://website.com/cts/api"
    body_xsl = "tests/test_data/xsl_test.xml"

    def setUp(self):
        with open("tests/test_data/getcapabilities.xml", "r") as f:
            self.getCapabilities = RequestPatch(f)

        with open("tests/test_data/getvalidreff.xml", "r") as f:
            self.getValidReff_single = RequestPatch(f)
            self.getValidReff = RequestPatchChained([self.getCapabilities, self.getValidReff_single])

        with open("tests/test_data/getpassage.xml", "r") as f:
            self.getPassage = RequestPatch(f)
            self.getPassage_Capabilities = RequestPatchChained([self.getCapabilities, self.getPassage])

        with open("tests/test_data/getprevnext.xml", "r") as f:
            self.getPrevNext = RequestPatch(f)
            self.getPassage_Route = RequestPatchChained([self.getCapabilities, self.getPassage, self.getPrevNext])

        self.nemo = Nemo(
            api_url=NemoResource.endpoint,
            app=Flask(__name__)
        )

NautilusDummy = NautilusRetriever(
    folders=[
        "./tests/test_data/nautilus/farsiLit",
        "./tests/test_data/nautilus/latinLit"
    ]
)
NautilusDummy.logger.setLevel(logging.ERROR)



"""

Test only controllers functions, ie function that should not be accessible directly from http routes but are still
greatly tied to the app/blueprint instance

"""
from flask.ext.nemo import Nemo
from mock import patch, call, Mock
import MyCapytain
from lxml import etree
from flask import Markup, Flask

from .resources import NemoResource


class NemoTestControllers(NemoResource):

    def test_flask_nemo(self):
        """ Testing Flask Nemo is set up"""
        a = Nemo()
        self.assertIsInstance(a, Nemo)
        a = Nemo()

    def test_without_inventory_request(self):
        """ Check that endpoint are derived from nemo.api_endpoint setting
        """
        #  Test without inventory
        with patch('requests.get', return_value=self.getCapabilities) as patched_get:
            self.nemo.get_inventory()
            patched_get.assert_called_once_with(
                NemoTestControllers.endpoint, params={
                    "request": "GetCapabilities"
                }
            )

    def test_with_inventory_request(self):
        """ Check that endpoint are derived from nemo.api_endpoint setting
        """
        with patch('requests.get', return_value=self.getCapabilities) as patched_get:
            #  Test with inventory
            self.nemo.api_inventory = "annotsrc"
            self.nemo.get_inventory()
            patched_get.assert_called_once_with(
                NemoTestControllers.endpoint, params={
                    "request": "GetCapabilities",
                    "inv": "annotsrc"
                }
            )
            self.nemo.api_inventory = None

    def test_inventory_parsing(self):
        """ Check that endpoint request leads to the creation of a TextInventory object
        """
        with patch('requests.get', return_value=self.getCapabilities) as patched_get:
            inventory = self.nemo.get_inventory()
            patched_get.assert_called_once_with(
                NemoTestControllers.endpoint, params={
                    "request": "GetCapabilities"
                }
            )
            self.assertIs(len(inventory.textgroups), 4)

    def test_get_collection(self):
        with patch('requests.get', return_value=self.getCapabilities) as patched_get:
            collections = self.nemo.get_collections()
            patched_get.assert_called_once_with(
                NemoTestControllers.endpoint, params={
                    "request": "GetCapabilities"
                }
            )
            self.assertEqual(collections, {"latinLit", "greekLit"})

    def test_get_authors(self):
        """ Check that authors textgroups are returned with informations
        """
        with patch('requests.get', return_value=self.getCapabilities):
            tgs = self.nemo.get_textgroups()
            self.assertIs(len(tgs), 4)
            self.assertEqual("urn:cts:greekLit:tlg0003" in [str(tg.urn) for tg in tgs], True)

    def test_get_authors_with_collection(self):
        """ Check that authors textgroups are returned when filtered by collection
        """
        with patch('requests.get', return_value=self.getCapabilities):
            tgs_2 = self.nemo.get_textgroups("greekLIT")  # With nice filtering
            self.assertIs(len(tgs_2), 1)
            self.assertEqual("urn:cts:greekLit:tlg0003" in [str(tg.urn) for tg in tgs_2], True)

    def test_get_works_with_collection(self):
        """ Check that works are returned when filtered by collection and textgroup
        """
        with patch('requests.get', return_value=self.getCapabilities):
            works = self.nemo.get_works("greekLIT", "TLG0003")  # With nice filtering
            self.assertIs(len(works), 1)
            self.assertEqual("urn:cts:greekLit:tlg0003.tlg001" in [str(work.urn) for work in works], True)

            #  Check when it fails
            works = self.nemo.get_works("greekLIT", "TLGabc003")  # With nice filtering
            self.assertIs(len(works), 0)

    def test_get_works_without_filters(self):
        """ Check that all works are returned when not filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            #  Check when it fails
            works = self.nemo.get_works()  # With nice filtering
            self.assertIs(len(works), 13)

    def test_get_works_with_one_filter(self):
        """ Check that error are raises
        """
        with self.assertRaises(ValueError):
            works = self.nemo.get_works("a", None)  # With nice filtering

        with self.assertRaises(ValueError):
            works = self.nemo.get_works(None, "a")

    def test_get_texts_with_all_filtered(self):
        """ Check that texts are filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            texts = self.nemo.get_texts("greekLIT", "TLG0003", "tlg001")  # With nice filtering
            self.assertIs(len(texts), 1)
            self.assertEqual("urn:cts:greekLit:tlg0003.tlg001.perseus-grc2" in [str(text.urn) for text in texts], True)

            texts = self.nemo.get_texts("greekLIT", "TLG0003", "tlg002")  # With nice filtering
            self.assertIs(len(texts), 0)

    def test_get_texts_with_none_filtered(self):
        """ Check that texts are filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            texts = self.nemo.get_texts()  # With nice filtering
            self.assertIs(len(texts), 14)
            self.assertEqual("urn:cts:greekLit:tlg0003.tlg001.perseus-grc2" in [str(text.urn) for text in texts], True)
            self.assertEqual("urn:cts:latinLit:phi1294.phi002.perseus-lat2" in [str(text.urn) for text in texts], True)
            self.assertEqual("urn:cts:latinLit:phi1294.phi002.perseus-lat3" in [str(text.urn) for text in texts], False)

    def test_get_texts_with_work_not_filtered(self):
        """ Check that all textgroups texts are returned
        """
        with patch('requests.get', return_value=self.getCapabilities):
            texts = self.nemo.get_texts("latinLit", "phi0959")  # With nice filtering
            self.assertIs(len(texts), 10)
            self.assertEqual("urn:cts:greekLit:tlg0003.tlg001.perseus-grc2" in [str(text.urn) for text in texts], False)
            self.assertEqual("urn:cts:latinLit:phi0959.tlg001.perseus-lat2" in [str(text.urn) for text in texts], False)

    def test_get_texts_raise(self):
        with self.assertRaises(
                ValueError):
            self.nemo.get_texts("latinLit")

    def test_get_single_text(self):
        """ Check that texts are filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            texts = self.nemo.get_text("greekLIT", "TLG0003", "tlg001", "perseus-grc2")  # With nice filtering
            self.assertIsInstance(texts, MyCapytain.resources.inventory.Text)
            self.assertEqual(str(texts.urn), "urn:cts:greekLit:tlg0003.tlg001.perseus-grc2")

    def test_get_single_text_empty_because_no_work(self):
        """ Check that texts are filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            with patch('flask_nemo.abort') as abort:
                texts = self.nemo.get_text("latinLit", "phi0959", "phi011", "perseus-lat2")   # With nice filtering
                abort.assert_called_once_with(404)

    def test_get_single_text_abort(self):
        """ Check that texts are filtered
        """
        with patch('requests.get', return_value=self.getCapabilities):
            with patch('flask_nemo.abort') as abort:
                texts = self.nemo.get_text("greekLIT", "TLG0003", "tlg001", "perseus-grc132")  # With nice filtering
                abort.assert_called_once_with(404)

    def test_get_validreffs_without_specific_callback(self):
        """ Try to get valid reffs
        """
        self.nemo = Nemo(api_url=NemoTestControllers.endpoint, inventory="annotsrc")
        with patch('requests.get', return_value=self.getValidReff) as patched:
            text, callback = self.nemo.get_reffs("latinLit", "phi1294", "phi002", "perseus-lat2")
            self.assertIsInstance(text, MyCapytain.resources.inventory.Text)

            reffs = callback(level=3)
            self.assertIsInstance(reffs, list)
            self.assertEqual(reffs[0], "urn:cts:latinLit:phi1294.phi002.perseus-lat2:1.pr.1")
            self.assertEqual(patched.mock_calls, [
                call(
                    NemoTestControllers.endpoint,
                    params={
                        "inv": "annotsrc",
                        "request": "GetCapabilities"
                    }
                ),
                call(
                    NemoTestControllers.endpoint,
                    params={
                        "inv": "annotsrc",
                        "request": "GetValidReff",
                        "level": "3",
                        "urn": "urn:cts:latinLit:phi1294.phi002.perseus-lat2"
                    }
                )
                ]
            )

    def test_get_passage(self):
        self.nemo = Nemo(api_url=NemoTestControllers.endpoint, inventory="annotsrc")
        with patch('requests.get', return_value=self.getPassage) as patched:
            passage = self.nemo.get_passage("latinLit", "phi1294", "phi002", "perseus-lat2", "1.pr")
            self.assertIsInstance(passage, MyCapytain.resources.texts.api.Passage)
            self.assertEqual(len(passage.xml.xpath("//tei:l[@n]", namespaces={"tei":"http://www.tei-c.org/ns/1.0"})), 6)
